@@ -28,39 +28,59 @@
  ****************************************************************************/
 
 
-static int Consumer_clear (Consumer *self) {
-	if (self->on_assign) {
-		Py_DECREF(self->on_assign);
-		self->on_assign = NULL;
+static int Consumer_clear (Handle *self) {
+	if (self->u.Consumer.on_assign) {
+		Py_DECREF(self->u.Consumer.on_assign);
+		self->u.Consumer.on_assign = NULL;
 	}
-	if (self->on_revoke) {
-		Py_DECREF(self->on_revoke);
-		self->on_revoke = NULL;
+	if (self->u.Consumer.on_revoke) {
+		Py_DECREF(self->u.Consumer.on_revoke);
+		self->u.Consumer.on_revoke = NULL;
 	}
-	if (self->on_commit) {
-		Py_DECREF(self->on_commit);
-		self->on_commit = NULL;
+	if (self->u.Consumer.on_commit) {
+		Py_DECREF(self->u.Consumer.on_commit);
+		self->u.Consumer.on_commit = NULL;
 	}
+
+	Handle_clear(self);
+
 	return 0;
 }
 
-static void Consumer_dealloc (Consumer *self) {
+static void Consumer_dealloc (Handle *self) {
 	PyObject_GC_UnTrack(self);
 
 	Consumer_clear(self);
 
-	if (self->rk)
-		rd_kafka_destroy(self->rk);
+	if (self->rk) {
+                CallState cs;
 
-	Py_TYPE(self)->tp_free((PyObject *)self);
+                CallState_begin(self, &cs);
+
+                /* If application has not called c.close() then
+                 * rd_kafka_destroy() will, and that might trigger
+                 * callbacks to be called from consumer_close().
+                 * This should probably be fixed in librdkafka,
+                 * or the application. */
+                rd_kafka_destroy(self->rk);
+
+                CallState_end(self, &cs);
+        }
+
+        Py_TYPE(self)->tp_free((PyObject *)self);
 }
 
-static int Consumer_traverse (Consumer *self,
-				  visitproc visit, void *arg) {
-	if (self->on_assign)
-		Py_VISIT(self->on_assign);
-	if (self->on_revoke)
-		Py_VISIT(self->on_revoke);
+static int Consumer_traverse (Handle *self,
+			      visitproc visit, void *arg) {
+	if (self->u.Consumer.on_assign)
+		Py_VISIT(self->u.Consumer.on_assign);
+	if (self->u.Consumer.on_revoke)
+		Py_VISIT(self->u.Consumer.on_revoke);
+	if (self->u.Consumer.on_commit)
+		Py_VISIT(self->u.Consumer.on_commit);
+
+	Handle_traverse(self, visit, arg);
+
 	return 0;
 }
 
@@ -69,7 +89,7 @@ static int Consumer_traverse (Consumer *self,
 
 
 
-static PyObject *Consumer_subscribe (Consumer *self, PyObject *args,
+static PyObject *Consumer_subscribe (Handle *self, PyObject *args,
 					 PyObject *kwargs) {
 
 	rd_kafka_topic_partition_list_t *topics;
@@ -100,7 +120,7 @@ static PyObject *Consumer_subscribe (Consumer *self, PyObject *args,
 		return NULL;
 	}
 
-	topics = rd_kafka_topic_partition_list_new(PyList_Size(tlist));
+	topics = rd_kafka_topic_partition_list_new((int)PyList_Size(tlist));
 	for (pos = 0 ; pos < PyList_Size(tlist) ; pos++) {
 		PyObject *o = PyList_GetItem(tlist, pos);
 		PyObject *uo;
@@ -130,29 +150,29 @@ static PyObject *Consumer_subscribe (Consumer *self, PyObject *args,
 	/*
 	 * Update rebalance callbacks
 	 */
-	if (self->on_assign) {
-		Py_DECREF(self->on_assign);
-		self->on_assign = NULL;
+	if (self->u.Consumer.on_assign) {
+		Py_DECREF(self->u.Consumer.on_assign);
+		self->u.Consumer.on_assign = NULL;
 	}
 	if (on_assign) {
-		self->on_assign = on_assign;
-		Py_INCREF(self->on_assign);
+		self->u.Consumer.on_assign = on_assign;
+		Py_INCREF(self->u.Consumer.on_assign);
 	}
 
-	if (self->on_revoke) {
-		Py_DECREF(self->on_revoke);
-		self->on_revoke = NULL;
+	if (self->u.Consumer.on_revoke) {
+		Py_DECREF(self->u.Consumer.on_revoke);
+		self->u.Consumer.on_revoke = NULL;
 	}
 	if (on_revoke) {
-		self->on_revoke = on_revoke;
-		Py_INCREF(self->on_revoke);
+		self->u.Consumer.on_revoke = on_revoke;
+		Py_INCREF(self->u.Consumer.on_revoke);
 	}
 
 	Py_RETURN_NONE;
 }
 
 
-static PyObject *Consumer_unsubscribe (Consumer *self,
+static PyObject *Consumer_unsubscribe (Handle *self,
 					   PyObject *ignore) {
 
 	rd_kafka_resp_err_t err;
@@ -169,7 +189,7 @@ static PyObject *Consumer_unsubscribe (Consumer *self,
 }
 
 
-static PyObject *Consumer_assign (Consumer *self, PyObject *tlist) {
+static PyObject *Consumer_assign (Handle *self, PyObject *tlist) {
 
 	rd_kafka_topic_partition_list_t *c_parts;
 	rd_kafka_resp_err_t err;
@@ -177,7 +197,7 @@ static PyObject *Consumer_assign (Consumer *self, PyObject *tlist) {
 	if (!(c_parts = py_to_c_parts(tlist)))
 		return NULL;
 
-	self->rebalance_assigned++;
+	self->u.Consumer.rebalance_assigned++;
 
 	err = rd_kafka_assign(self->rk, c_parts);
 
@@ -185,7 +205,7 @@ static PyObject *Consumer_assign (Consumer *self, PyObject *tlist) {
 
 	if (err) {
 		cfl_PyErr_Format(err,
-				 "Failed to set assignemnt: %s",
+				 "Failed to set assignment: %s",
 				 rd_kafka_err2str(err));
 		return NULL;
 	}
@@ -194,11 +214,11 @@ static PyObject *Consumer_assign (Consumer *self, PyObject *tlist) {
 }
 
 
-static PyObject *Consumer_unassign (Consumer *self, PyObject *ignore) {
+static PyObject *Consumer_unassign (Handle *self, PyObject *ignore) {
 
 	rd_kafka_resp_err_t err;
 
-	self->rebalance_assigned++;
+	self->u.Consumer.rebalance_assigned++;
 
 	err = rd_kafka_assign(self->rk, NULL);
 	if (err) {
@@ -211,9 +231,31 @@ static PyObject *Consumer_unassign (Consumer *self, PyObject *ignore) {
 	Py_RETURN_NONE;
 }
 
+static PyObject *Consumer_assignment (Handle *self, PyObject *args,
+                                      PyObject *kwargs) {
+
+        PyObject *plist;
+        rd_kafka_topic_partition_list_t *c_parts;
+        rd_kafka_resp_err_t err;
+
+        err = rd_kafka_assignment(self->rk, &c_parts);
+        if (err) {
+                cfl_PyErr_Format(err,
+                                 "Failed to get assignment: %s",
+                                 rd_kafka_err2str(err));
+                return NULL;
+        }
 
 
-static PyObject *Consumer_commit (Consumer *self, PyObject *args,
+        plist = c_parts_to_py(c_parts);
+        rd_kafka_topic_partition_list_destroy(c_parts);
+
+        return plist;
+}
+
+
+
+static PyObject *Consumer_commit (Handle *self, PyObject *args,
 					PyObject *kwargs) {
 
 	rd_kafka_resp_err_t err;
@@ -259,7 +301,6 @@ static PyObject *Consumer_commit (Consumer *self, PyObject *args,
 
 	} else {
 		c_offsets = NULL;
-		err = rd_kafka_commit(self->rk, NULL, async);
 	}
 
 			
@@ -281,7 +322,7 @@ static PyObject *Consumer_commit (Consumer *self, PyObject *args,
 
 
 
-static PyObject *Consumer_committed (Consumer *self, PyObject *args,
+static PyObject *Consumer_committed (Handle *self, PyObject *args,
 					 PyObject *kwargs) {
 
 	PyObject *plist;
@@ -317,7 +358,7 @@ static PyObject *Consumer_committed (Consumer *self, PyObject *args,
 }
 
 
-static PyObject *Consumer_position (Consumer *self, PyObject *args,
+static PyObject *Consumer_position (Handle *self, PyObject *args,
 					PyObject *kwargs) {
 
 	PyObject *plist;
@@ -351,43 +392,97 @@ static PyObject *Consumer_position (Consumer *self, PyObject *args,
 }
 
 
+static PyObject *Consumer_get_watermark_offsets (Handle *self, PyObject *args,
+                                                 PyObject *kwargs) {
 
-static PyObject *Consumer_poll (Consumer *self, PyObject *args,
-				    PyObject *kwargs) {
-	double tmout = -1.0f;
-	static char *kws[] = { "timeout", NULL };
-	rd_kafka_message_t *rkm;
-	PyObject *msgobj;
+        TopicPartition *tp;
+        rd_kafka_resp_err_t err;
+        double tmout = -1.0f;
+        int cached = 0;
+        int64_t low = RD_KAFKA_OFFSET_INVALID, high = RD_KAFKA_OFFSET_INVALID;
+        static char *kws[] = { "partition", "timeout", "cached", NULL };
+        PyObject *rtup;
 
-	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|d", kws, &tmout))
-		return NULL;
+        if (!PyArg_ParseTupleAndKeywords(args, kwargs, "O|db", kws,
+                                         (PyObject **)&tp, &tmout, &cached))
+                return NULL;
 
-	self->callback_crashed = 0;
-	self->thread_state = PyEval_SaveThread();
 
-	rkm = rd_kafka_consumer_poll(self->rk, tmout >= 0 ?
-				     (int)(tmout * 1000.0f) : -1);
+        if (PyObject_Type((PyObject *)tp) != (PyObject *)&TopicPartitionType) {
+                PyErr_Format(PyExc_TypeError,
+                             "expected %s", TopicPartitionType.tp_name);
+                return NULL;
+        }
 
-	PyEval_RestoreThread(self->thread_state);
-	self->thread_state = NULL;
+        if (cached) {
+                err = rd_kafka_get_watermark_offsets(self->rk,
+                                                     tp->topic, tp->partition,
+                                                     &low, &high);
+        } else {
+                err = rd_kafka_query_watermark_offsets(self->rk,
+                                                       tp->topic, tp->partition,
+                                                       &low, &high,
+                                                       tmout >= 0 ? (int)(tmout * 1000.0f) : -1);
+        }
 
-	if (self->callback_crashed)
-		return NULL;
+        if (err) {
+                cfl_PyErr_Format(err,
+                                 "Failed to get watermark offsets: %s",
+                                 rd_kafka_err2str(err));
+                return NULL;
+        }
 
-	if (!rkm)
-		Py_RETURN_NONE;
+        rtup = PyTuple_New(2);
+        PyTuple_SetItem(rtup, 0, PyLong_FromLongLong(low));
+        PyTuple_SetItem(rtup, 1, PyLong_FromLongLong(high));
 
-	msgobj = Message_new0(rkm);
-	rd_kafka_message_destroy(rkm);
-
-	return msgobj;
+        return rtup;
 }
 
 
-static PyObject *Consumer_close (Consumer *self, PyObject *ignore) {
-	self->thread_state = PyEval_SaveThread();
+
+static PyObject *Consumer_poll (Handle *self, PyObject *args,
+                                    PyObject *kwargs) {
+        double tmout = -1.0f;
+        static char *kws[] = { "timeout", NULL };
+        rd_kafka_message_t *rkm;
+        PyObject *msgobj;
+        CallState cs;
+
+        if (!PyArg_ParseTupleAndKeywords(args, kwargs, "|d", kws, &tmout))
+                return NULL;
+
+        CallState_begin(self, &cs);
+
+        rkm = rd_kafka_consumer_poll(self->rk, tmout >= 0 ?
+                                     (int)(tmout * 1000.0f) : -1);
+
+        if (!CallState_end(self, &cs))
+                return NULL;
+
+        if (!rkm)
+                Py_RETURN_NONE;
+
+        msgobj = Message_new0(rkm);
+        rd_kafka_message_destroy(rkm);
+
+        return msgobj;
+}
+
+
+static PyObject *Consumer_close (Handle *self, PyObject *ignore) {
+	CallState cs;
+
+	CallState_begin(self, &cs);
+
 	rd_kafka_consumer_close(self->rk);
-	PyEval_RestoreThread(self->thread_state);
+
+	rd_kafka_destroy(self->rk);
+	self->rk = NULL;
+
+        if (!CallState_end(self, &cs))
+                return NULL;
+
 	Py_RETURN_NONE;
 }
 
@@ -399,6 +494,11 @@ static PyMethodDef Consumer_methods[] = {
 	  "\n"
 	  "  Set subscription to supplied list of topics\n"
 	  "  This replaces a previous subscription.\n"
+          "\n"
+          "  Regexp pattern subscriptions are supported by prefixing "
+          "the topic string with ``\"^\"``, e.g.::\n"
+          "\n"
+          "    consumer.subscribe([\"^my_topic.*\", \"^another[0-9]-?[a-z]+$\", \"not_a_regex\"])\n"
 	  "\n"
 	  "  :param list(str) topics: List of topics (strings) to subscribe to.\n"
 	  "  :param callable on_assign: callback to provide handling of "
@@ -418,10 +518,11 @@ static PyMethodDef Consumer_methods[] = {
 	  "  :param list(TopicPartition) partitions: Absolute list of partitions being assigned or revoked.\n"
 	  "\n"
 	},
-	{ "unsubscribe", (PyCFunction)Consumer_unsubscribe, METH_NOARGS,
-	  "  Remove current subscription.\n"
-	  "\n"
-	},
+        { "unsubscribe", (PyCFunction)Consumer_unsubscribe, METH_NOARGS,
+          "  Remove current subscription.\n"
+          "  :raises: KafkaException\n"
+          "\n"
+        },
 	{ "poll", (PyCFunction)Consumer_poll,
 	  METH_VARARGS|METH_KEYWORDS,
 	  ".. py:function:: poll([timeout=None])\n"
@@ -450,17 +551,30 @@ static PyMethodDef Consumer_methods[] = {
 	  "  :param list(TopicPartition) partitions: List of topic+partitions and optionally initial offsets to start consuming.\n"
 	  "\n"
 	},
-	{ "unassign", (PyCFunction)Consumer_unassign, METH_NOARGS,
-	  "  Removes the current partition assignment and stops consuming.\n"
-	  "\n"
-	},
+        { "unassign", (PyCFunction)Consumer_unassign, METH_NOARGS,
+          "  Removes the current partition assignment and stops consuming.\n"
+          "  :raises: KafkaException\n"
+          "\n"
+        },
+        { "assignment", (PyCFunction)Consumer_assignment,
+          METH_VARARGS|METH_KEYWORDS,
+          ".. py:function:: assignment()\n"
+          "\n"
+          "  Returns the current partition assignment.\n"
+          "\n"
+          "  :returns: List of assigned topic+partitions.\n"
+          "  :rtype: list(TopicPartition)\n"
+          "  :raises: KafkaException\n"
+          "\n"
+        },
 	{ "commit", (PyCFunction)Consumer_commit, METH_VARARGS|METH_KEYWORDS,
 	  ".. py:function:: commit([message=None], [offsets=None], [async=True])\n"
 	  "\n"
 	  "  Commit a message or a list of offsets.\n"
 	  "\n"
 	  "  ``message`` and ``offsets`` are mutually exclusive, if neither is set "
-	  "the current partition assignment's offsets are used instead.\n"
+	  "the current partition assignment's offsets are used instead. "
+	  "The consumer relies on your use of this method if you have set 'enable.auto.commit' to False\n"
 	  "\n"
 	  "  :param confluent_kafka.Message message: Commit message's offset+1.\n"
 	  "  :param list(TopicPartition) offsets: List of topic+partitions+offsets to commit.\n"
@@ -497,6 +611,22 @@ static PyMethodDef Consumer_methods[] = {
 	  "  :raises: KafkaException\n"
 	  "\n"
 	},
+        { "get_watermark_offsets", (PyCFunction)Consumer_get_watermark_offsets,
+          METH_VARARGS|METH_KEYWORDS,
+          ".. py:function:: get_watermark_offsets(partition, [timeout=None], [cached=False])\n"
+          "\n"
+          "  Retrieve low and high offsets for partition.\n"
+          "\n"
+          "  :param TopicPartition partition: Topic+partition to return offsets for."
+          "  :param float timeout: Request timeout (when cached=False).\n"
+          "  :param bool cached: Instead of querying the broker used cached information. "
+          "Cached values: The low offset is updated periodically (if statistics.interval.ms is set) while "
+          "the high offset is updated on each message fetched from the broker for this partition."
+          "  :returns: Tuple of (low,high) on success or None on timeout.\n"
+          "  :rtype: tuple(int,int)\n"
+          "  :raises: KafkaException\n"
+          "\n"
+        },
 	{ "close", (PyCFunction)Consumer_close, METH_NOARGS,
 	  "\n"
 	  "  Close down and terminate the Kafka Consumer.\n"
@@ -504,7 +634,7 @@ static PyMethodDef Consumer_methods[] = {
 	  "  Actions performed:\n"
 	  "\n"
 	  "  - Stops consuming\n"
-	  "  - Commits offsets\n"
+	  "  - Commits offsets - except if the consumer property 'enable.auto.commit' is set to False\n"
 	  "  - Leave consumer group\n"
 	  "\n"
 	  "  .. note: Registered callbacks may be called from this method, "
@@ -517,13 +647,160 @@ static PyMethodDef Consumer_methods[] = {
 };
 
 
+static void Consumer_rebalance_cb (rd_kafka_t *rk, rd_kafka_resp_err_t err,
+				   rd_kafka_topic_partition_list_t *c_parts,
+				   void *opaque) {
+	Handle *self = opaque;
+	CallState *cs;
+
+	cs = CallState_get(self);
+
+	self->u.Consumer.rebalance_assigned = 0;
+
+	if ((err == RD_KAFKA_RESP_ERR__ASSIGN_PARTITIONS &&
+	     self->u.Consumer.on_assign) ||
+	    (err == RD_KAFKA_RESP_ERR__REVOKE_PARTITIONS &&
+	     self->u.Consumer.on_revoke)) {
+		PyObject *parts;
+		PyObject *args, *result;
+
+		/* Construct list of TopicPartition based on 'c_parts' */
+		parts = c_parts_to_py(c_parts);
+
+		args = Py_BuildValue("(OO)", self, parts);
+
+		Py_DECREF(parts);
+
+		if (!args) {
+			cfl_PyErr_Format(RD_KAFKA_RESP_ERR__FAIL,
+					 "Unable to build callback args");
+			CallState_crash(cs);
+			CallState_resume(cs);
+			return;
+		}
+
+		result = PyObject_CallObject(
+			err == RD_KAFKA_RESP_ERR__ASSIGN_PARTITIONS ?
+			self->u.Consumer.on_assign :
+			self->u.Consumer.on_revoke, args);
+
+		Py_DECREF(args);
+
+		if (result)
+			Py_DECREF(result);
+		else {
+			CallState_crash(cs);
+			rd_kafka_yield(rk);
+		}
+	}
+
+	/* Fallback: librdkafka needs the rebalance_cb to call assign()
+	 * to synchronize state, if the user did not do this from callback,
+	 * or there was no callback, or the callback failed, then we perform
+	 * that assign() call here instead. */
+	if (!self->u.Consumer.rebalance_assigned) {
+		if (err == RD_KAFKA_RESP_ERR__ASSIGN_PARTITIONS)
+			rd_kafka_assign(rk, c_parts);
+		else
+			rd_kafka_assign(rk, NULL);
+	}
+
+	CallState_resume(cs);
+}
+
+
+static void Consumer_offset_commit_cb (rd_kafka_t *rk, rd_kafka_resp_err_t err,
+				       rd_kafka_topic_partition_list_t *c_parts,
+				       void *opaque) {
+	Handle *self = opaque;
+	PyObject *parts, *k_err, *args, *result;
+	CallState *cs;
+
+	if (!self->u.Consumer.on_commit)
+		return;
+
+	cs = CallState_get(self);
+
+	/* Insantiate error object */
+	k_err = KafkaError_new_or_None(err, NULL);
+
+	/* Construct list of TopicPartition based on 'c_parts' */
+	if (c_parts)
+		parts = c_parts_to_py(c_parts);
+	else
+		parts = PyList_New(0);
+
+	args = Py_BuildValue("(OO)", k_err, parts);
+
+	Py_DECREF(k_err);
+	Py_DECREF(parts);
+
+	if (!args) {
+		cfl_PyErr_Format(RD_KAFKA_RESP_ERR__FAIL,
+				 "Unable to build callback args");
+		CallState_crash(cs);
+		CallState_resume(cs);
+		return;
+	}
+
+	result = PyObject_CallObject(self->u.Consumer.on_commit, args);
+
+	Py_DECREF(args);
+
+	if (result)
+		Py_DECREF(result);
+	else {
+		CallState_crash(cs);
+		rd_kafka_yield(rk);
+	}
+
+	CallState_resume(cs);
+}
+
+
+
+static int Consumer_init (PyObject *selfobj, PyObject *args, PyObject *kwargs) {
+        Handle *self = (Handle *)selfobj;
+        char errstr[256];
+        rd_kafka_conf_t *conf;
+
+        if (self->rk) {
+                PyErr_SetString(PyExc_RuntimeError,
+                                "Consumer already __init__:ialized");
+                return -1;
+        }
+
+        if (!(conf = common_conf_setup(RD_KAFKA_CONSUMER, self,
+                                       args, kwargs)))
+                return -1; /* Exception raised by ..conf_setup() */
+
+        rd_kafka_conf_set_rebalance_cb(conf, Consumer_rebalance_cb);
+        rd_kafka_conf_set_offset_commit_cb(conf, Consumer_offset_commit_cb);
+
+        self->rk = rd_kafka_new(RD_KAFKA_CONSUMER, conf,
+                                errstr, sizeof(errstr));
+        if (!self->rk) {
+                cfl_PyErr_Format(rd_kafka_last_error(),
+                                 "Failed to create consumer: %s", errstr);
+                rd_kafka_conf_destroy(conf);
+                return -1;
+        }
+
+        rd_kafka_poll_set_consumer(self->rk);
+
+        return 0;
+}
+
 static PyObject *Consumer_new (PyTypeObject *type, PyObject *args,
-				   PyObject *kwargs);
+                               PyObject *kwargs) {
+        return type->tp_alloc(type, 0);
+}
+
 
 PyTypeObject ConsumerType = {
 	PyVarObject_HEAD_INIT(NULL, 0)
 	"cimpl.Consumer",        /*tp_name*/
-	sizeof(Consumer),      /*tp_basicsize*/
+	sizeof(Handle),          /*tp_basicsize*/
 	0,                         /*tp_itemsize*/
 	(destructor)Consumer_dealloc, /*tp_dealloc*/
 	0,                         /*tp_print*/
@@ -540,7 +817,8 @@ PyTypeObject ConsumerType = {
 	0,                         /*tp_getattro*/
 	0,                         /*tp_setattro*/
 	0,                         /*tp_as_buffer*/
-	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC, /*tp_flags*/
+	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE |
+	Py_TPFLAGS_HAVE_GC, /*tp_flags*/
 	"High-level Kafka Consumer\n"
 	"\n"
 	".. py:function:: Consumer(**kwargs)\n"
@@ -574,148 +852,7 @@ PyTypeObject ConsumerType = {
 	0,                         /* tp_descr_get */
 	0,                         /* tp_descr_set */
 	0,                         /* tp_dictoffset */
-	0,                         /* tp_init */
+        Consumer_init,             /* tp_init */
 	0,                         /* tp_alloc */
 	Consumer_new           /* tp_new */
 };
-
-
-
-static void Consumer_rebalance_cb (rd_kafka_t *rk, rd_kafka_resp_err_t err,
-				   rd_kafka_topic_partition_list_t *c_parts,
-				   void *opaque) {
-	Consumer *self = opaque;
-
-	PyEval_RestoreThread(self->thread_state);
-
-	self->rebalance_assigned = 0;
-
-	if ((err == RD_KAFKA_RESP_ERR__ASSIGN_PARTITIONS && self->on_assign) ||
-	    (err == RD_KAFKA_RESP_ERR__REVOKE_PARTITIONS && self->on_revoke)) {
-		PyObject *parts;
-		PyObject *args, *result;
-
-		/* Construct list of TopicPartition based on 'c_parts' */
-		parts = c_parts_to_py(c_parts);
-
-		args = Py_BuildValue("(OO)", self, parts);
-
-		Py_DECREF(parts);
-
-		if (!args) {
-			cfl_PyErr_Format(RD_KAFKA_RESP_ERR__FAIL,
-					 "Unable to build callback args");
-			self->thread_state = PyEval_SaveThread();
-			self->callback_crashed++;
-			return;
-		}
-
-		result = PyObject_CallObject(
-			err == RD_KAFKA_RESP_ERR__ASSIGN_PARTITIONS ?
-			self->on_assign : self->on_revoke, args);
-
-		Py_DECREF(args);
-
-		if (result)
-			Py_DECREF(result);
-		else {
-			self->callback_crashed++;
-			rd_kafka_yield(rk);
-		}
-	}
-
-	/* Fallback: librdkafka needs the rebalance_cb to call assign()
-	 * to synchronize state, if the user did not do this from callback,
-	 * or there was no callback, or the callback failed, then we perform
-	 * that assign() call here instead. */
-	if (!self->rebalance_assigned) {
-		if (err == RD_KAFKA_RESP_ERR__ASSIGN_PARTITIONS)
-			rd_kafka_assign(rk, c_parts);
-		else
-			rd_kafka_assign(rk, NULL);
-	}
-
-	self->thread_state = PyEval_SaveThread();
-}
-
-
-static void Consumer_offset_commit_cb (rd_kafka_t *rk, rd_kafka_resp_err_t err,
-				       rd_kafka_topic_partition_list_t *c_parts,
-				       void *opaque) {
-	Consumer *self = opaque;
-	PyObject *parts, *k_err, *args, *result;
-
-	if (!self->on_commit)
-		return;
-
-	PyEval_RestoreThread(self->thread_state);
-
-	/* Insantiate error object */
-	k_err = KafkaError_new_or_None(err, NULL);
-
-	/* Construct list of TopicPartition based on 'c_parts' */
-	parts = c_parts_to_py(c_parts);
-
-	args = Py_BuildValue("(OO)", k_err, parts);
-
-	Py_DECREF(k_err);
-	Py_DECREF(parts);
-
-	if (!args) {
-		cfl_PyErr_Format(RD_KAFKA_RESP_ERR__FAIL,
-				 "Unable to build callback args");
-		self->thread_state = PyEval_SaveThread();
-		self->callback_crashed++;
-		return;
-	}
-
-	result = PyObject_CallObject(self->on_commit, args);
-
-	Py_DECREF(args);
-
-	if (result)
-		Py_DECREF(result);
-	else {
-		self->callback_crashed++;
-		rd_kafka_yield(rk);
-	}
-
-	self->thread_state = PyEval_SaveThread();
-}
-
-
-
-static PyObject *Consumer_new (PyTypeObject *type, PyObject *args,
-				   PyObject *kwargs) {
-	Consumer *self;
-	char errstr[256];
-	rd_kafka_conf_t *conf;
-
-	self = (Consumer *)ConsumerType.tp_alloc(&ConsumerType, 0);
-	if (!self)
-		return NULL;
-
-	if (!(conf = common_conf_setup(RD_KAFKA_CONSUMER, self,
-					   args, kwargs))) {
-		Py_DECREF(self);
-		return NULL;
-	}
-
-	rd_kafka_conf_set_rebalance_cb(conf, Consumer_rebalance_cb);
-	rd_kafka_conf_set_offset_commit_cb(conf, Consumer_offset_commit_cb);
-
-	self->rk = rd_kafka_new(RD_KAFKA_CONSUMER, conf,
-				errstr, sizeof(errstr));
-	if (!self->rk) {
-		cfl_PyErr_Format(rd_kafka_last_error(),
-				 "Failed to create consumer: %s", errstr);
-		Py_DECREF(self);
-		return NULL;
-	}
-
-	rd_kafka_poll_set_consumer(self->rk);
-
-	return (PyObject *)self;
-}
-
-
