@@ -29,8 +29,6 @@
  *     partitioner run without any locks taken.
  *     Until this is fixed the partitioner is ignored and librdkafka's
  *     default will be used.
- *  - TopicPartition.offset should probably be None for the INVALID offset
- *    rather than exposing the special value -1001.
  *  - KafkaError type .tp_doc allocation is lost on exit.
  *
  */
@@ -133,7 +131,7 @@ static PyObject* KafkaError_richcompare (KafkaError *self, PyObject *o2,
 	if (Py_TYPE(o2) == &KafkaErrorType)
 		code2 = ((KafkaError *)o2)->code;
 	else
-		code2 = PyLong_AsLong(o2);
+		code2 = (int)PyLong_AsLong(o2);
 
 	switch (op)
 	{
@@ -164,7 +162,7 @@ static PyObject* KafkaError_richcompare (KafkaError *self, PyObject *o2,
 	Py_INCREF(result);
 	return result;
 }
-				       
+
 
 static PyTypeObject KafkaErrorType = {
 	PyVarObject_HEAD_INIT(NULL, 0)
@@ -183,10 +181,10 @@ static PyTypeObject KafkaErrorType = {
 	(hashfunc)KafkaError_hash, /*tp_hash */
 	0,                         /*tp_call*/
 	0,                         /*tp_str*/
-	0,                         /*tp_getattro*/
+	PyObject_GenericGetAttr,   /*tp_getattro*/
 	0,                         /*tp_setattro*/
 	0,                         /*tp_as_buffer*/
-	Py_TPFLAGS_DEFAULT,        /*tp_flags*/
+	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE, /*tp_flags*/
 	"Kafka error and event object\n"
 	"\n"
 	"  The KafkaError class serves multiple purposes:\n"
@@ -320,12 +318,36 @@ static PyObject *Message_partition (Message *self, PyObject *ignore) {
 
 static PyObject *Message_offset (Message *self, PyObject *ignore) {
 	if (self->offset >= 0)
-		return PyLong_FromLong(self->offset);
+		return PyLong_FromLongLong(self->offset);
 	else
 		Py_RETURN_NONE;
 }
 
 
+static PyObject *Message_timestamp (Message *self, PyObject *ignore) {
+	return Py_BuildValue("iL",
+			     self->tstype,
+			     self->timestamp);
+}
+
+static PyObject *Message_set_value (Message *self, PyObject *new_val) {
+   if (self->value)
+        Py_DECREF(self->value);
+   self->value = new_val;
+   Py_INCREF(self->value);
+
+   Py_RETURN_NONE;
+}
+
+static PyObject *Message_set_key (Message *self, PyObject *new_key) {
+   if (self->key)
+        Py_DECREF(self->key);
+   self->key = new_key;
+   Py_INCREF(self->key);
+
+   Py_RETURN_NONE;
+}
+
 static PyMethodDef Message_methods[] = {
 	{ "error", (PyCFunction)Message_error, METH_NOARGS,
 	  "  The message object is also used to propagate errors and events, "
@@ -362,6 +384,42 @@ static PyMethodDef Message_methods[] = {
 	  "  :rtype: int or None\n"
 	  "\n"
 	},
+	{ "timestamp", (PyCFunction)Message_timestamp, METH_NOARGS,
+          "  Retrieve timestamp type and timestamp from message.\n"
+          "  The timestamp type is one of:\n"
+          "    * :py:const:`TIMESTAMP_NOT_AVAILABLE`"
+          " - Timestamps not supported by broker\n"
+          "    * :py:const:`TIMESTAMP_CREATE_TIME` "
+          " - Message creation time (or source / producer time)\n"
+          "    * :py:const:`TIMESTAMP_LOG_APPEND_TIME` "
+          " - Broker receive time\n"
+          "\n"
+          "  The returned timestamp should be ignored if the timestamp type is "
+          ":py:const:`TIMESTAMP_NOT_AVAILABLE`.\n"
+          "\n"
+          "  The timestamp is the number of milliseconds since the epoch (UTC).\n"
+          "\n"
+          "  Timestamps require broker version 0.10.0.0 or later and \n"
+          "  ``{'api.version.request': True}`` configured on the client.\n"
+          "\n"
+	  "  :returns: tuple of message timestamp type, and timestamp.\n"
+	  "  :rtype: (int, int)\n"
+	  "\n"
+	},
+	{ "set_value", (PyCFunction)Message_set_value, METH_O,
+	  "  Set the field 'Message.value' with new value.\n"
+	  "  :param: object value: Message.value.\n"
+	  "  :returns: None.\n"
+	  "  :rtype: None\n"
+	  "\n"
+	},
+	{ "set_key", (PyCFunction)Message_set_key, METH_O,
+	  "  Set the field 'Message.key' with new value.\n"
+	  "  :param: object value: Message.key.\n"
+	  "  :returns: None.\n"
+	  "  :rtype: None\n"
+	  "\n"
+	},
 	{ NULL }
 };
 
@@ -434,14 +492,15 @@ PyTypeObject MessageType = {
 	0,                         /*tp_getattro*/
 	0,                         /*tp_setattro*/
 	0,                         /*tp_as_buffer*/
-	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC, /*tp_flags*/
+	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE |
+	Py_TPFLAGS_HAVE_GC, /*tp_flags*/
 	"The Message object represents either a single consumed or "
 	"produced message, or an event (:py:func:`error()` is not None).\n"
 	"\n"
 	"An application must check with :py:func:`error()` to see if the "
 	"object is a proper message (error() returns None) or an "
 	"error/event.\n"
-	"\n"
+        "\n"
 	"This class is not user-instantiable.\n"
 	"\n"
 	".. py:function:: len()\n"
@@ -495,6 +554,8 @@ PyObject *Message_new0 (const rd_kafka_message_t *rkm) {
 	self->partition = rkm->partition;
 	self->offset = rkm->offset;
 
+	self->timestamp = rd_kafka_message_timestamp(rkm, &self->tstype);
+
 	return (PyObject *)self;
 }
 
@@ -510,15 +571,6 @@ PyObject *Message_new0 (const rd_kafka_message_t *rkm) {
  *
  *
  ****************************************************************************/
-typedef struct {
-	PyObject_HEAD
-	char *topic;
-	int   partition;
-	int64_t offset;
-	PyObject *error;
-} TopicPartition;
-
-
 static int TopicPartition_clear (TopicPartition *self) {
 	if (self->topic) {
 		free(self->topic);
@@ -531,6 +583,16 @@ static int TopicPartition_clear (TopicPartition *self) {
 	return 0;
 }
 
+static void TopicPartition_setup (TopicPartition *self, const char *topic,
+				  int partition, long long offset,
+				  rd_kafka_resp_err_t err) {
+	self->topic = strdup(topic);
+	self->partition = partition;
+	self->offset = offset;
+	self->error = KafkaError_new_or_None(err, NULL);
+}
+
+
 static void TopicPartition_dealloc (TopicPartition *self) {
 	PyObject_GC_UnTrack(self);
 
@@ -539,13 +601,9 @@ static void TopicPartition_dealloc (TopicPartition *self) {
 	Py_TYPE(self)->tp_free((PyObject *)self);
 }
 
-static PyObject *TopicPartition_new0 (const char *topic, int partition,
-				      long long offset,
-				      rd_kafka_resp_err_t err);
 
-
-static PyObject *TopicPartition_new (PyTypeObject *type, PyObject *args,
-				     PyObject *kwargs) {
+static int TopicPartition_init (PyObject *self, PyObject *args,
+				      PyObject *kwargs) {
 	const char *topic;
 	int partition = RD_KAFKA_PARTITION_UA;
 	long long offset = RD_KAFKA_OFFSET_INVALID;
@@ -556,11 +614,21 @@ static PyObject *TopicPartition_new (PyTypeObject *type, PyObject *args,
 
 	if (!PyArg_ParseTupleAndKeywords(args, kwargs, "s|iL", kws,
 					 &topic, &partition, &offset))
-		return NULL;
+		return -1;
 
-	return TopicPartition_new0(topic, partition, offset, 0);
+	TopicPartition_setup((TopicPartition *)self,
+			     topic, partition, offset, 0);
+
+	return 0;
 }
-	
+
+
+static PyObject *TopicPartition_new (PyTypeObject *type, PyObject *args,
+				     PyObject *kwargs) {
+	PyObject *self = type->tp_alloc(type, 1);
+	return self;
+}
+
 
 
 static int TopicPartition_traverse (TopicPartition *self,
@@ -572,15 +640,21 @@ static int TopicPartition_traverse (TopicPartition *self,
 
 
 static PyMemberDef TopicPartition_members[] = {
-	{ "topic", T_STRING, offsetof(TopicPartition, topic), READONLY,
-	  "Topic name" },
-	{ "partition", T_INT, offsetof(TopicPartition, partition), 0,
-	  "Partition number" },
-	{ "offset", T_LONGLONG, offsetof(TopicPartition, offset), 0,
-	  "Offset" }, /* FIXME: Possibly use None for INVALID offset (-1001) */
-	{ "error", T_OBJECT, offsetof(TopicPartition, error), READONLY,
-	  "Indicates an error (with :py:class:`KafkaError`) unless None." },
-	{ NULL }
+        { "topic", T_STRING, offsetof(TopicPartition, topic), READONLY,
+          ":py:attribute:topic - Topic name (string)" },
+        { "partition", T_INT, offsetof(TopicPartition, partition), 0,
+          ":py:attribute: Partition number (int)" },
+        { "offset", T_LONGLONG, offsetof(TopicPartition, offset), 0,
+          " :py:attribute: Offset (long)\n"
+          "Either an absolute offset (>=0) or a logical offset:"
+          " :py:const:`OFFSET_BEGINNING`,"
+          " :py:const:`OFFSET_END`,"
+          " :py:const:`OFFSET_STORED`,"
+          " :py:const:`OFFSET_INVALID`"
+        },
+        { "error", T_OBJECT, offsetof(TopicPartition, error), READONLY,
+          ":py:attribute: Indicates an error (with :py:class:`KafkaError`) unless None." },
+        { NULL }
 };
 
 
@@ -588,11 +662,13 @@ static PyObject *TopicPartition_str0 (TopicPartition *self) {
 	PyObject *errstr = self->error == Py_None ? NULL :
 		cfl_PyObject_Unistr(self->error);
 	PyObject *ret;
+	char offset_str[40];
+	snprintf(offset_str, sizeof(offset_str), "%"PRId64"", self->offset);
 	ret = cfl_PyUnistr(
 		_FromFormat("TopicPartition{topic=%s,partition=%"PRId32
-			    ",offset=%"PRId64",error=%s}",
+			    ",offset=%s,error=%s}",
 			    self->topic, self->partition,
-			    self->offset,
+			    offset_str,
 			    errstr ? cfl_PyUnistr_AsUTF8(errstr) : "None"));
 	if (errstr)
 		Py_DECREF(errstr);
@@ -600,8 +676,6 @@ static PyObject *TopicPartition_str0 (TopicPartition *self) {
 }
 
 
-static PyTypeObject TopicPartitionType;
-
 static PyObject *
 TopicPartition_richcompare (TopicPartition *self, PyObject *o2,
 			    int op) {
@@ -658,7 +732,7 @@ static long TopicPartition_hash (TopicPartition *self) {
 }
 
 
-static PyTypeObject TopicPartitionType = {
+PyTypeObject TopicPartitionType = {
 	PyVarObject_HEAD_INIT(NULL, 0)
 	"cimpl.TopicPartition",         /*tp_name*/
 	sizeof(TopicPartition),       /*tp_basicsize*/
@@ -675,10 +749,11 @@ static PyTypeObject TopicPartitionType = {
 	(hashfunc)TopicPartition_hash, /*tp_hash */
 	0,                         /*tp_call*/
 	0,                         /*tp_str*/
-	0,                         /*tp_getattro*/
+	PyObject_GenericGetAttr,   /*tp_getattro*/
 	0,                         /*tp_setattro*/
 	0,                         /*tp_as_buffer*/
-	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC, /*tp_flags*/
+	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE |
+	Py_TPFLAGS_HAVE_GC, /*tp_flags*/
 	"TopicPartition is a generic type to hold a single partition and "
 	"various information about it.\n"
 	"\n"
@@ -709,9 +784,9 @@ static PyTypeObject TopicPartitionType = {
 	0,                         /* tp_descr_get */
 	0,                         /* tp_descr_set */
 	0,                         /* tp_dictoffset */
-	0,                         /* tp_init */
+	TopicPartition_init,       /* tp_init */
 	0,                         /* tp_alloc */
-	TopicPartition_new     /* tp_new */
+	TopicPartition_new         /* tp_new */
 };
 
 /**
@@ -722,15 +797,10 @@ static PyObject *TopicPartition_new0 (const char *topic, int partition,
 				      rd_kafka_resp_err_t err) {
 	TopicPartition *self;
 
-	self = (TopicPartition *)TopicPartitionType.tp_alloc(
-		&TopicPartitionType, 0);
-	if (!self)
-		return NULL;
+	self = (TopicPartition *)TopicPartitionType.tp_new(
+		&TopicPartitionType, NULL, NULL);
 
-	self->topic = strdup(topic);
-	self->partition = partition;
-	self->offset = offset;
-	self->error = KafkaError_new_or_None(err, NULL);
+	TopicPartition_setup(self, topic, partition, offset, err);
 
 	return (PyObject *)self;
 }
@@ -758,7 +828,7 @@ PyObject *c_parts_to_py (const rd_kafka_topic_partition_list_t *c_parts) {
 	}
 
 	return parts;
-		
+
 }
 
 /**
@@ -776,7 +846,7 @@ rd_kafka_topic_partition_list_t *py_to_c_parts (PyObject *plist) {
 		return NULL;
 	}
 
-	c_parts = rd_kafka_topic_partition_list_new(PyList_Size(plist));
+	c_parts = rd_kafka_topic_partition_list_new((int)PyList_Size(plist));
 
 	for (i = 0 ; i < PyList_Size(plist) ; i++) {
 		TopicPartition *tp = (TopicPartition *)
@@ -801,7 +871,67 @@ rd_kafka_topic_partition_list_t *py_to_c_parts (PyObject *plist) {
 }
 
 
+/****************************************************************************
+ *
+ *
+ * Common callbacks
+ *
+ *
+ *
+ *
+ ****************************************************************************/
+static void error_cb (rd_kafka_t *rk, int err, const char *reason, void *opaque) {
+	Handle *h = opaque;
+	PyObject *eo, *result;
+	CallState *cs;
 
+	cs = CallState_get(h);
+	if (!h->error_cb) {
+		/* No callback defined */
+		goto done;
+	}
+
+	eo = KafkaError_new0(err, "%s", reason);
+	result = PyObject_CallFunctionObjArgs(h->error_cb, eo, NULL);
+	Py_DECREF(eo);
+
+	if (result)
+		Py_DECREF(result);
+	else {
+		CallState_crash(cs);
+		rd_kafka_yield(h->rk);
+	}
+
+ done:
+	CallState_resume(cs);
+}
+
+static int stats_cb(rd_kafka_t *rk, char *json, size_t json_len, void *opaque) {
+	Handle *h = opaque;
+	PyObject *eo = NULL, *result = NULL;
+	CallState *cs = NULL;
+
+	cs = CallState_get(h);
+	if (json_len == 0) {
+		/* No data returned*/
+		goto done;
+	}
+
+	eo = Py_BuildValue("s", json);
+	result = PyObject_CallFunctionObjArgs(h->stats_cb, eo, NULL);
+	Py_DECREF(eo);
+
+	if (result)
+		Py_DECREF(result);
+	else {
+		CallState_crash(cs);
+		rd_kafka_yield(h->rk);
+	}
+
+ done:
+	CallState_resume(cs);
+	return 0;
+}
 
 /****************************************************************************
  *
@@ -814,6 +944,36 @@ rd_kafka_topic_partition_list_t *py_to_c_parts (PyObject *plist) {
  ****************************************************************************/
 
 
+
+/**
+ * Clear Python object references in Handle
+ */
+void Handle_clear (Handle *h) {
+	if (h->error_cb)
+		Py_DECREF(h->error_cb);
+
+	if (h->stats_cb)
+		Py_DECREF(h->stats_cb);
+
+        if (h->initiated)
+                PyThread_delete_key(h->tlskey);
+}
+
+/**
+ * GC traversal for Python object references
+ */
+int Handle_traverse (Handle *h, visitproc visit, void *arg) {
+	if (h->error_cb)
+		Py_VISIT(h->error_cb);
+
+	if (h->stats_cb)
+		Py_VISIT(h->stats_cb);
+
+	return 0;
+}
+
+
+
 /**
  * Populate topic conf from provided dict.
  *
@@ -879,7 +1039,7 @@ static int populate_topic_conf (rd_kafka_topic_conf_t *tconf, const char *what,
  *
  * @returns 1 if handled, 0 if unknown, or -1 on failure (exception raised).
  */
-static int producer_conf_set_special (Producer *self, rd_kafka_conf_t *conf,
+static int producer_conf_set_special (Handle *self, rd_kafka_conf_t *conf,
 				      rd_kafka_topic_conf_t *tconf,
 				      const char *name, PyObject *valobj) {
 	PyObject *vs;
@@ -894,8 +1054,8 @@ static int producer_conf_set_special (Producer *self, rd_kafka_conf_t *conf,
 			return -1;
 		}
 
-		self->default_dr_cb = valobj;
-		Py_INCREF(self->default_dr_cb);
+		self->u.Producer.default_dr_cb = valobj;
+		Py_INCREF(self->u.Producer.default_dr_cb);
 
 		return 1;
 
@@ -947,11 +1107,11 @@ static int producer_conf_set_special (Producer *self, rd_kafka_conf_t *conf,
 				return -1;
 			}
 
-			if (self->partitioner_cb)
-				Py_DECREF(self->partitioner_cb);
+			if (self->u.Producer.partitioner_cb)
+				Py_DECREF(self->u.Producer.partitioner_cb);
 
-			self->partitioner_cb = valobj;
-			Py_INCREF(self->partitioner_cb);
+			self->u.Producer.partitioner_cb = valobj;
+			Py_INCREF(self->u.Producer.partitioner_cb);
 
 			/* Use trampoline to call Python code. */
 			rd_kafka_topic_conf_set_partitioner_cb(tconf,
@@ -970,7 +1130,7 @@ static int producer_conf_set_special (Producer *self, rd_kafka_conf_t *conf,
  *
  * @returns 1 if handled, 0 if unknown, or -1 on failure (exception raised).
  */
-static int consumer_conf_set_special (Consumer *self, rd_kafka_conf_t *conf,
+static int consumer_conf_set_special (Handle *self, rd_kafka_conf_t *conf,
 				      rd_kafka_topic_conf_t *tconf,
 				      const char *name, PyObject *valobj) {
 
@@ -983,8 +1143,8 @@ static int consumer_conf_set_special (Consumer *self, rd_kafka_conf_t *conf,
 			return -1;
 		}
 
-		self->on_commit = valobj;
-		Py_INCREF(self->on_commit);
+		self->u.Consumer.on_commit = valobj;
+		Py_INCREF(self->u.Consumer.on_commit);
 
 		return 1;
 	}
@@ -1000,7 +1160,7 @@ static int consumer_conf_set_special (Consumer *self, rd_kafka_conf_t *conf,
  * an exception has been raised.
  */
 rd_kafka_conf_t *common_conf_setup (rd_kafka_type_t ktype,
-				    void *self0,
+				    Handle *h,
 				    PyObject *args,
 				    PyObject *kwargs) {
 	rd_kafka_conf_t *conf;
@@ -1053,15 +1213,55 @@ rd_kafka_conf_t *common_conf_setup (rd_kafka_type_t ktype,
 
 			Py_DECREF(ks);
 			continue;
+
+		} else if (!strcmp(k, "error_cb")) {
+			if (!PyCallable_Check(vo)) {
+				PyErr_SetString(PyExc_TypeError,
+						"expected error_cb property "
+						"as a callable function");
+				rd_kafka_topic_conf_destroy(tconf);
+				rd_kafka_conf_destroy(conf);
+				Py_DECREF(ks);
+				return NULL;
+                        }
+			if (h->error_cb) {
+				Py_DECREF(h->error_cb);
+				h->error_cb = NULL;
+			}
+			if (vo != Py_None) {
+				h->error_cb = vo;
+				Py_INCREF(h->error_cb);
+			}
+			Py_DECREF(ks);
+			continue;
+		} else if (!strcmp(k, "stats_cb")) {
+			if (!PyCallable_Check(vo)) {
+				PyErr_SetString(PyExc_TypeError,
+						"expected stats_cb property "
+						"as a callable function");
+				rd_kafka_topic_conf_destroy(tconf);
+				rd_kafka_conf_destroy(conf);
+				Py_DECREF(ks);
+				return NULL;
+                        }
+
+			if (h->stats_cb) {
+				Py_DECREF(h->stats_cb);
+				h->stats_cb = NULL;
+			}
+			if (vo != Py_None) {
+				h->stats_cb = vo;
+				Py_INCREF(h->stats_cb);
+			}
+			Py_DECREF(ks);
+			continue;
 		}
 
 		/* Special handling for certain config keys. */
 		if (ktype == RD_KAFKA_PRODUCER)
-			r = producer_conf_set_special((Producer *)self0,
-						      conf, tconf, k, vo);
+			r = producer_conf_set_special(h, conf, tconf, k, vo);
 		else
-			r = consumer_conf_set_special((Consumer *)self0,
-						      conf, tconf, k, vo);
+			r = consumer_conf_set_special(h, conf, tconf, k, vo);
 		if (r == -1) {
 			/* Error */
 			Py_DECREF(ks);
@@ -1104,10 +1304,19 @@ rd_kafka_conf_t *common_conf_setup (rd_kafka_type_t ktype,
 		Py_DECREF(ks);
 	}
 
-	rd_kafka_topic_conf_set_opaque(tconf, self0);
+	if (h->error_cb)
+		rd_kafka_conf_set_error_cb(conf, error_cb);
+
+	if (h->stats_cb)
+		rd_kafka_conf_set_stats_cb(conf, stats_cb);
+
+	rd_kafka_topic_conf_set_opaque(tconf, h);
 	rd_kafka_conf_set_default_topic_conf(conf, tconf);
 
-	rd_kafka_conf_set_opaque(conf, self0);
+	rd_kafka_conf_set_opaque(conf, h);
+
+	h->tlskey = PyThread_create_key();
+        h->initiated = 1;
 
 	return conf;
 }
@@ -1115,6 +1324,61 @@ rd_kafka_conf_t *common_conf_setup (rd_kafka_type_t ktype,
 
 
 
+/**
+ * @brief Initialiase a CallState and unlock the GIL prior to a
+ *        possibly blocking external call.
+ */
+void CallState_begin (Handle *h, CallState *cs) {
+	cs->thread_state = PyEval_SaveThread();
+	assert(cs->thread_state != NULL);
+	cs->crashed = 0;
+	PyThread_set_key_value(h->tlskey, cs);
+}
+
+/**
+ * @brief Relock the GIL after external call is done.
+ * @returns 0 if a Python signal was raised or a callback crashed, else 1.
+ */
+int CallState_end (Handle *h, CallState *cs) {
+	PyThread_delete_key_value(h->tlskey);
+
+	PyEval_RestoreThread(cs->thread_state);
+
+	if (PyErr_CheckSignals() == -1 || cs->crashed)
+		return 0;
+
+	return 1;
+}
+
+
+/**
+ * @brief Get the current thread's CallState and re-locks the GIL.
+ */
+CallState *CallState_get (Handle *h) {
+	CallState *cs = PyThread_get_key_value(h->tlskey);
+	assert(cs != NULL);
+	assert(cs->thread_state != NULL);
+	PyEval_RestoreThread(cs->thread_state);
+	cs->thread_state = NULL;
+	return cs;
+}
+
+/**
+ * @brief Un-locks the GIL to resume blocking external call.
+ */
+void CallState_resume (CallState *cs) {
+	assert(cs->thread_state == NULL);
+	cs->thread_state = PyEval_SaveThread();
+}
+
+/**
+ * @brief Indicate that call crashed.
+ */
+void CallState_crash (CallState *cs) {
+	cs->crashed++;
+}
+
+
 
 /****************************************************************************
  *
@@ -1134,7 +1398,7 @@ static PyObject *libversion (PyObject *self, PyObject *args) {
 }
 
 static PyObject *version (PyObject *self, PyObject *args) {
-	return Py_BuildValue("si", "0.9.1", 0x00090100);
+	return Py_BuildValue("si", "0.9.2", 0x00090100);
 }
 
 static PyMethodDef cimpl_methods[] = {
@@ -1294,6 +1558,15 @@ static PyObject *_init_cimpl (void) {
 	Py_INCREF(KafkaException);
 	PyModule_AddObject(m, "KafkaException", KafkaException);
 
+	PyModule_AddIntConstant(m, "TIMESTAMP_NOT_AVAILABLE", RD_KAFKA_TIMESTAMP_NOT_AVAILABLE);
+	PyModule_AddIntConstant(m, "TIMESTAMP_CREATE_TIME", RD_KAFKA_TIMESTAMP_CREATE_TIME);
+	PyModule_AddIntConstant(m, "TIMESTAMP_LOG_APPEND_TIME", RD_KAFKA_TIMESTAMP_LOG_APPEND_TIME);
+
+        PyModule_AddIntConstant(m, "OFFSET_BEGINNING", RD_KAFKA_OFFSET_BEGINNING);
+        PyModule_AddIntConstant(m, "OFFSET_END", RD_KAFKA_OFFSET_END);
+        PyModule_AddIntConstant(m, "OFFSET_STORED", RD_KAFKA_OFFSET_STORED);
+        PyModule_AddIntConstant(m, "OFFSET_INVALID", RD_KAFKA_OFFSET_INVALID);
+
 	return m;
 }
 
